# Copyright (c) 2012, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import tempfile
import yaml
import urllib2

import rospkg.distro
import rosdep2.sources_list

def get_test_dir():
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'sources.list.d'))

def test_url_constants():
    from rosdep2.rep3 import REP3_TARGETS_URL
    for url_name, url in [('REP3_TARGETS_URL', REP3_TARGETS_URL),
                          ]:
        try:
            f = urllib2.urlopen(url)
            f.read()
            f.close()
        except:
            assert False, "URL [%s][%s] failed to download"%(url_name, url)

def test_download_targets_data():
    from rosdep2.rep3 import download_targets_data, REP3_TARGETS_URL
    from rosdep2 import DownloadFailure
    data = download_targets_data(REP3_TARGETS_URL)
    assert type(data) is dict
    assert 'electric' in data
    assert 'fuerte' in data
    assert 'lucid' in data['fuerte']
    assert 'oneiric' in data['fuerte']

    try:
        download_targets_data(targets_url='http://bad.ros.org/foo.yaml')
        assert False, "should have raised"
    except DownloadFailure:
        pass



import os
import sys
import urllib2
import yaml

from rospkg.os_detect import OS_UBUNTU

from .core import InvalidData, DownloadFailure
from .platforms.debian import APT_INSTALLER
from .rep3 import download_targets_data

#py3k
try:
    unicode
except:
    basestring = unicode = str

# location of an example gbpdistro file for reference and testing
FUERTE_GBPDISTRO_URL = 'https://raw.github.com/ros/rosdistro/master/releases/fuerte.yaml'

#seconds to wait before aborting download of gbpdistro data
DOWNLOAD_TIMEOUT = 15.0 

def gbprepo_to_rosdep_data(gbpdistro_data, targets_data):
    """
    :raises: :exc:`InvalidData`
    """
    # Error reporting for this isn't nearly as good as it could be
    # (e.g. doesn't separate gbpdistro vs. targets, nor provide
    # origin), but rushing this implementation a bit.
    try:
        if not type(targets_data) == dict:
            raise InvalidData("targets data must be a dict")
        if not type(gbpdistro_data) == dict:
            raise InvalidData("gbpdistro data must be a dictionary")        

        # compute the default target data for the release_name
        release_name = gbpdistro_data['release-name']
        if not release_name in targets_data:
            raise InvalidData("targets file does not contain information for release [%s]"%(release_name))
        else:
            # take the first match
            target_data = targets_data[release_name]

        # compute the rosdep data for each repo
        rosdep_data = {}
        gbp_repos = gbpdistro_data['gbp-repos']
        for repo in gbp_repos:
            if type(repo) != dict:
                raise InvalidData("invalid repo spec in gbpdistro data: %s"%(str(repo)))
            rosdep_key = repo['name']
            rosdep_data[rosdep_key] = {}

            # Do generation for ubuntu
            rosdep_data[rosdep_key][OS_UBUNTU] = {}

            # - debian package name: underscores must be dashes
            deb_package_name = 'ros-%s-%s'%(release_name, rosdep_key)
            deb_package_name = deb_package_name.replace('_', '-')

            repo_targets = repo['target']
            if repo_targets == 'all':
                repo_targets = target_data

            for t in repo_targets:
                if not isinstance(t, basestring):
                    raise InvalidData("invalid target spec: %s"%(t))
                rosdep_data[rosdep_key][OS_UBUNTU][t] = {APT_INSTALLER:
                                                         {'packages': [deb_package_name]}}
        return rosdep_data
    except KeyError as e:
        raise InvalidData("Invalid GBP-distro/targets format: missing key: %s"%(str(e)))
    
def download_gbpdistro_as_rosdep_data(gbpdistro_url, targets_url=None):
    """
    Download gbpdistro file from web and convert format to rosdep distro data.
    
    :param gbpdistro_url: url of gbpdistro file, ``str``
    :param target_url: override URL of platform targets file
    :raises: :exc:`DownloadFailure`
    :raises: :exc:`InvalidData` If targets file does not pass cursory validation checks.
    """
    # we can convert a gbpdistro file into rosdep data by following a couple rules
    targets_data = download_targets_data(targets_url=targets_url)
    try:
        f = urllib2.urlopen(gbpdistro_url, timeout=DOWNLOAD_TIMEOUT)
        text = f.read()
        f.close()
        gbpdistro_data = yaml.safe_load(text)
        return gbprepo_to_rosdep_data(gbpdistro_data, targets_data)
    except Exception as e:
        raise DownloadFailure("Failed to download target platform data for gbpdistro:\n\t%s"%(str(e)))




"""
Helper routines for catkin.  These are distributed inside of rosdep2
to protect catkin against future rosdep2 API updatese.  These helper
routines are assumed to run in an interactive mode with an end-user
and thus return end-user oriented error messages.

Errors are returned as arguments to raised :exc:`ValidationFailed`
exceptions.

Workflow::

    installer = get_apt_installer()
    view = get_catkin_view(rosdistro_name, 'ubuntu', 'lucid')
    resolve_for_apt(rosdep_key, view, installer, 'ubuntu', 'lucid')

"""

from __future__ import print_function

import os
import sys

from subprocess import Popen, PIPE, CalledProcessError

from . import create_default_installer_context
from .platforms.debian import APT_INSTALLER
from .rep3 import download_targets_data
from .sources_list import get_sources_list_dir, DataSourceMatcher, SourcesListLoader
from .lookup import RosdepLookup
from .rospkg_loader import DEFAULT_VIEW_KEY

class ValidationFailed(Exception):
    pass

def call(command, pipe=None):
    """
    Copy of call() function from catkin-generate-debian to mimic output
    """
    working_dir = '.'
    #print('+ cd %s && ' % working_dir + ' '.join(command))
    process = Popen(command, stdout=pipe, stderr=pipe, cwd=working_dir)
    output, unused_err = process.communicate()
    retcode = process.poll()
    if retcode:
        raise CalledProcessError(retcode, command)
    if pipe:
        return output

def get_ubuntu_targets(rosdistro):
    """
    Get a list of Ubuntu distro codenames for the specified ROS
    distribution.  This method blocks on an HTTP download.

    :raises: :exc:`ValidationFailed`
    """
    targets_data = download_targets_data()
    return targets_data[rosdistro]

def get_apt_installer():
    installer_context = create_default_installer_context()
    return installer_context.get_installer(APT_INSTALLER)

def resolve_for_apt(rosdep_key, view, installer, os_name, os_version):
    """
    Resolve rosdep key to apt dependencies.
    
    :param os_name: OS name, e.g. 'ubuntu'

    :raises: :exc:`rosdep2.ResolutionError`
    """
    d = view.lookup(rosdep_key)
    inst_key, rule = d.get_rule_for_platform(os_name, os_version, [APT_INSTALLER], APT_INSTALLER)    
    assert inst_key == APT_INSTALLER
    return installer.resolve(rule)

def get_catkin_view(rosdistro_name, os_name, os_version):
    """
    :raises: :exc:`ValidationFailed`
    """
    sources_list_dir = get_sources_list_dir()
    if not os.path.exists(sources_list_dir):
        raise ValidationFailed("""rosdep database is not initialized, please run:
\tsudo rosdep init
""")
    # may want to make the update optional
    call(('rosdep', 'update'), pipe=PIPE)

    sources_matcher = DataSourceMatcher([rosdistro_name, os_name, os_version])
    sources_loader = SourcesListLoader.create_default(matcher=sources_matcher)
    if not (sources_loader.sources):
        raise ValidationFailed("""rosdep database does not have any sources.
Please make sure you have a valid configuration in:
\t%s
"""%(sources_list_dir))
    
    # for vestigial reasons, using the roskg loader, but we're only
    # actually using the backend db as resolution is not resource-name based
    lookup = RosdepLookup.create_from_rospkg(sources_loader=sources_loader)
    return lookup.get_rosdep_view(DEFAULT_VIEW_KEY)



# Copyright (c) 2011, Willow Garage, Inc.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the Willow Garage, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Author Ken Conley/kwc@willowgarage.com, Murph Finnicum/murph@murph.cc

import os
import traceback
from mock import Mock, patch
from rosdep2.model import InvalidData

def get_test_dir():
    # not used yet
    return os.path.abspath(os.path.join(os.path.dirname(__file__), 'gentoo'))

from rospkg.os_detect import OsDetect, OS_GENTOO

def test_portage_available():
    from rosdep2.platforms.gentoo import portage_available

    original_exists = os.path.exists

    path_overrides = {}
    def mock_path(path):
        if path in path_overrides:
            return path_overrides[path]
        else: 
            return original_exists(path)

    m = Mock(side_effect=mock_path)
    os.path.exists = m

    #Test with portageq missing
    m.reset_mock()
    path_overrides = {}
    path_overrides['/usr/bin/portageq'] = False
    path_overrides['/usr/bin/emerge'] = True

    val = portage_available()
    assert val==False, "Portage should not be available without portageq"

    #Test with emerge missing
    m.reset_mock()
    path_overrides = {}
    path_overrides['/usr/bin/portageq'] = True
    path_overrides['/usr/bin/emerge'] = False

    val = portage_available()
    assert val==False, "Portage should not be available without emerge"

    # Test with nothing missing
    m.reset_mock()
    path_overrides = {}
    path_overrides['/usr/bin/portageq'] = True
    path_overrides['/usr/bin/emerge'] = True

    val = portage_available()
    assert val==True, "Portage should be available"
    
    os.path.exists = original_exists

# This actually tests portage_detect_single and portage_detect

def test_portage_detect():
    from rosdep2.platforms.gentoo import portage_detect

    m = Mock()
    m.return_value = []

    val = portage_detect([], exec_fn=m)
    assert val == [], val
    
    # Test checking for a package that we do not have installed
    m = Mock(return_value = [])
    val = portage_detect(['tinyxml[stl]'], exec_fn=m)
    assert val == [], "Result was actually: %s" % val
    m.assert_called_with(['portageq', 'match', '/', 'tinyxml[stl]'])

    # Test checking for a package that we do have installed
    m = Mock(return_value = ['dev-libs/tinyxml-2.6.2-r1'])
    val = portage_detect(['tinyxml[stl]'], exec_fn=m)
    assert val == ['tinyxml[stl]'], "Result was actually: %s" % val
    m.assert_called_with(['portageq', 'match', '/', 'tinyxml[stl]'])
    
    # Test checking for two packages that we have installed
    m = Mock(side_effect = [['sys-devel/gcc-4.5.3-r2'], ['dev-libs/tinyxml-2.6.2-r1']])
    val = portage_detect(['tinyxml[stl]', 'gcc'], exec_fn=m)
    assert val == ['gcc', 'tinyxml[stl]'], "Result was actually: %s" % val
    m.assert_any_call(['portageq', 'match', '/', 'tinyxml[stl]'])   
    m.assert_any_call(['portageq', 'match', '/', 'gcc'])
    
    # Test checking for two missing packages
    m = Mock(side_effect = [[],[]])

    val = portage_detect(['tinyxml[stl]', 'gcc'], exec_fn=m)
    assert val == [], "Result was actually: %s" % val
    m.assert_any_call(['portageq', 'match', '/', 'tinyxml[stl]'])   
    m.assert_any_call(['portageq', 'match', '/', 'gcc'])

    # Test checking for one missing, one installed package
    m = Mock(side_effect = [['sys-devel/gcc-4.5.3-r2'], []])

    val = portage_detect(['tinyxml[stl]', 'gcc'], exec_fn=m)
    assert val == ['gcc'], "Result was actually: %s" % val
    m.assert_any_call(['portageq', 'match', '/', 'tinyxml[stl]'])   
    m.assert_any_call(['portageq', 'match', '/', 'gcc'])

    # Test checking for one installed, one missing package (reverse order)
    m = Mock(side_effect = [[], ['dev-libs/tinyxml-2.6.2-r1']])

    val = portage_detect(['tinyxml[stl]', 'gcc'], exec_fn=m)
    assert val == ['tinyxml[stl]'], "Result was actually: %s" % val
    m.assert_any_call(['portageq', 'match', '/', 'tinyxml[stl]'])   
    m.assert_any_call(['portageq', 'match', '/', 'gcc'])

    # Test duplicates (requesting the same package twice)
    #TODO what's the desired behavior here
    m = Mock(side_effect = [['dev-libs/tinyxml-2.6.2-r1'],['dev-libs/tinyxml-2.6.2-r1']])

    val = portage_detect(['tinyxml[stl]', 'tinyxml[stl]'], exec_fn=m)
    assert val == ['tinyxml[stl]','tinyxml[stl]'], "Result was actually: %s" % val
    m.assert_any_call(['portageq', 'match', '/', 'tinyxml[stl]'])   
    # and a second of the same, but any_call won't show that.

    # Test packages with multiple slot
    m = Mock(side_effect = [['dev-lang/python-2.7.2-r3','dev-lang/python-3.2.2']])
    val = portage_detect(['python'], exec_fn=m)
    assert val == ['python'], "Result was actually: %s" % val
    m.assert_any_call(['portageq', 'match', '/', 'python'])   

def test_PortageInstaller():
    from rosdep2.platforms.gentoo import PortageInstaller

    @patch.object(PortageInstaller, 'get_packages_to_install')
    def test(mock_method):
        installer = PortageInstaller()
        mock_method.return_value = []
        assert [] == installer.get_install_command(['fake'])

        mock_method.return_value = ['a', 'b']
        
        expected = [['sudo', 'emerge', 'a'],
                    ['sudo', 'emerge', 'b']]
        val = installer.get_install_command(['whatever'], interactive=False)
        assert val == expected, val

        expected = [['sudo', 'emerge', '-a', 'a'],
                    ['sudo', 'emerge', '-a', 'b']]
        val = installer.get_install_command(['whatever'], interactive=True)
        assert val == expected, val
        
    try:
        test()
    except AssertionError:
        traceback.print_exc()
        raise
    



from rosdep2.catkin_support import get_apt_installer, get_catkin_view, resolve_for_apt, ValidationFailed

def test_workflow():
    try:
        installer = get_apt_installer()
        view = get_catkin_view('fuerte', 'ubuntu', 'lucid')
        resolved = resolve_for_apt('cmake', view, installer, 'ubuntu', 'lucid')
        assert ['cmake'] == resolved
        resolved = resolve_for_apt('python', view, installer, 'ubuntu', 'lucid')
        assert resolved == ['python-dev']
    except ValidationFailed:
        # tests fail on the server because 'rosdep init' has not been run
        pass
